#!/usr/bin/env ruby -wKU
#
#  First Cuts
#
#  Created by Gregory Clarke on 2008-04-05.
#  Copyright (c) 2008. All rights reserved.
#

require 'serendipity_algorithm.bundle'
$first_cuts_studio = true
require 'exceLogger_algorithm.bundle'

class AppController < OSX::NSObject

	ib_outlets :mainWindow, :deactivateMenuItem, :getInfoMenuItem, :saveSnapshotMenuItem
	ib_outlets :firstCutsView, :firstCutsSpinner, :infoTextField, :firstCutsBrowseFileButton, :durationRadioButtons, :durationTextField, :openerPopup, :firstCutsLowerThirdPopup, :frameRatePopup, :frameSizePopup, :aspectRatioPopup, :anamorphicCheckbox, :randomizeButton, :keywordsAvailableTableView, :keywordsChosenTableView, :addButton, :removeButton, :removeAllButton, :previewArollButton, :startTheEditButton
	ib_outlets :progressWindow, :progressView, :progressIndicator, :progressMessage
	ib_outlets :finisherView, :finishTheEditButton, :finisherBrowseFileButton, :finisherLowerThirdPopup, :finisherSpinner
	ib_outlets :exceLoggerView, :chooseExcelButton, :excelIcon, :excelFileName, :chooseXMLButton, :xmlIcon, :xmlFileName, :saveXMLButton, :exceLoggerSpinner, :onSaveRadioButtons
	ib_outlets :regoView, :deactivateAreaView, :webView, :problemMessage, :closeRegoButton

	# String constants for the toolbar.
	TOOLBAR_IDENTIFIER = 'My Window Toolbar Identifier'
	TOOLBAR_FIRST_IDENTIFIER = 'First Toolbar Item Identifier'
	TOOLBAR_SECOND_IDENTIFIER = 'Second Toolbar Item Identifier'
	TOOLBAR_THIRD_IDENTIFIER = 'Third Toolbar Item Identifier'
	# Pasteboards for drag-and-drop table views
	KeywordsAvailablePasteboard = OSX::NSString.stringWithString "KeywordsAvailablePasteboard"
	KeywordsChosenPasteboard = OSX::NSString.stringWithString "KeywordsChosenPasteboard"

	ib_action :showFirstCuts do |sender|
		showFirstCutsView()
	end

	ib_action :showFinisher do |sender|
		showFinisherView()
	end
	
	ib_action :showExceLogger do |sender|
		showExceLoggerView()
	end

	def initialize
		@story_keywords_available = Array.new
		@columns = ["minutes", "keyword"]
		@story_keywords_chosen = Array.new
		@excel_file = nil
		@xml_file = nil
	end
	
	def awakeFromNib
		puts "awakeFromNib"
		setupToolbar
		@keywordsChosenTableView.registerForDraggedTypes([KeywordsAvailablePasteboard])
		@keywordsChosenTableView.registerForDraggedTypes([KeywordsChosenPasteboard])
		@keywordsAvailableTableView.registerForDraggedTypes([KeywordsChosenPasteboard])
		# will use CoreAnimation for the window changes:
		@mainWindow.contentView.wantsLayer = true
		# check for Application Support folder
		paths = OSX::NSSearchPathForDirectoriesInDomains(OSX::NSApplicationSupportDirectory, OSX::NSUserDomainMask, true)
		if paths.length > 0
			ENV['ROO_TMP'] = paths[0].stringByExpandingTildeInPath.to_s
		end
	end

	### rego methods ###
	
	def applicationWillFinishLaunching(sender) # applicationWillFinishLaunching: is sent before application:openFile: (so it can deal with a file dropped onto the app icon)
		puts "applicationWillFinishLaunching"
		check_for_expired_beta()
		if @this_rego[0]["active"]
			showFirstCutsView()
			@deactivateMenuItem.setEnabled(true)
		elsif @this_rego[0]["mac_problem"]==true
			check_for_rego(@this_rego)
		else
			displayDemoSheet()
		end
	end
	
	def displayDemoSheet
		alert = OSX::NSAlert.alloc.init
		alert.setMessageText("#{SKU} has not been activated for this computer.")
		alert.setAlertStyle(OSX::NSInformationalAlertStyle)
		alert.setInformativeText("Would you like to activate #{SKU}, or use it in demo mode (which is unable to save or export)?")
		alert.addButtonWithTitle("Activate")
		alert.addButtonWithTitle("Buy")
		alert.addButtonWithTitle("Demo")
		alert.beginSheetModalForWindow_modalDelegate_didEndSelector_contextInfo(@mainWindow, self, "didEndDemoSheet:returnCode:contextInfo:", nil)
	end

	def didEndDemoSheet_returnCode_contextInfo(sheet, buttonClicked, info)
		sheet.window.orderOut(nil)
		if buttonClicked==OSX::NSAlertSecondButtonReturn
			system('open', 'http://www.AssistedEditing.com/FirstCuts/buy')
		end
		if buttonClicked==OSX::NSAlertThirdButtonReturn
			showFirstCutsView()
		else
			newFrame = self.newFrameForNewContentView(@regoView)
			showActivationView(newFrame, WEB_APP + "license/" + SKU.gsub(" ", "%20") + "?sn=" + @this_rego[0]["number"].to_s + "&mac=" + MAC) # NEW
			@deactivateMenuItem.setEnabled(false)
		end
	end

	def showDeactivation
		@toolbar.setVisible(false)
		newFrame = self.newFrameForNewContentView(@deactivateAreaView) # 500 x 200
		@mainWindow.setMinSize(OSX::NSSize.new(500, 200+22))
		@mainWindow.setMaxSize(OSX::NSSize.new(500, 200+22))
		@mainWindow.setContentView(@deactivateAreaView)
		showActivationView(newFrame, WEB_APP + "destroy/" + @this_rego[0]["number"].to_s)
		@closeRegoButton.setTitle("Cancel")
	end
	ib_action :showDeactivation # First Cuts > Deactivate
	
	def showActivationView(frame, url)
		puts "showActivationView #{url}"
		@toolbar.setVisible(false)
		@mainWindow.setFrame_display frame, true
		@mainWindow.setContentView(@regoView)
		@mainWindow.setTitle("Activation")
		# notification
		center = OSX::NSNotificationCenter.defaultCenter
		center.addObserver_selector_name_object(self, :webViewProgressFinished, "WebProgressFinishedNotification", @webView) # name is WebProgressFinishedNotification NOT WebViewProgressFinishedNotification
		@webView.mainFrame.loadRequest( OSX::NSURLRequest.requestWithURL( OSX::NSURL.URLWithString(url) ) )
	end
	
	def webViewProgressFinished(notification)
		puts notification.name
		if notification.name=="WebProgressFinishedNotification"
			if @webView.mainFrame.dataSource.nil? # in case page doesn't load...
				@webView.mainFrame.stopLoading
				@problemMessage.setHidden(false)
				@webView.setHidden(true)
			else
				get_web_app_response(@webView.mainFrame.dataSource.response.URL)
			end
		end
	end 
	
	def closeActivationView
		if @closeRegoButton.title=="Continue" || @closeRegoButton.title=="Cancel"
			@toolbar.setVisible(true)
			showFirstCutsView()
		else
			OSX::NSApplication.sharedApplication.terminate(self)
		end
	end
	ib_action :closeActivationView # on click of Registration window's Continue button
	
	### end rego methods ###
	
	def showFirstCutsView
		puts "showFirstCutsView"
		@toolbar.selectedItemIdentifier = @toolbar.items[0].itemIdentifier # select the first toolbar item
		setupView(@firstCutsView, "First Cuts", 700, 600)
	end
	
	def showFinisherView
		puts "showFinisherView"
		@toolbar.selectedItemIdentifier = @toolbar.items[1].itemIdentifier
		setupView(@finisherView, "Finisher", 500, 140)
	end

	def showExceLoggerView
		puts "showExceLoggerView"
		setupView(@exceLoggerView, "exceLogger", 500, 280)
	end
	
	def setupView(view, title, width, height)
		@mainWindow.setTitle(title)
		oldFrame = @mainWindow.frame
		newFrame = self.newFrameForNewContentView(view)
		@mainWindow.setMinSize(OSX::NSSize.new(width, height+22))
		@mainWindow.setMaxSize(OSX::NSSize.new(width, height+22))
		@mainWindow.setFrame_display_animate(newFrame, true, true)
		@mainWindow.setContentView(view)
	end

	def newFrameForNewContentView(view)
		newFrameRect = @mainWindow.frameRectForContentRect(view.frame)
		oldFrameRect = @mainWindow.frame
		newSize = newFrameRect.size
		oldSize = oldFrameRect.size
		frame = @mainWindow.frame
		frame.size = newSize
		frame.origin.y = frame.origin.y - (newSize.height - oldSize.height)
		frame
	end

	def showAbout(sender)
		OSX::NSApp.orderFrontStandardAboutPanelWithOptions(nil)
		# OSX::NSApp.orderFrontStandardAboutPanelWithOptions({ "Copyright" => "\n\nhttp://www.AssistedEditing.com\n\nCopyright © 2008 Intelligent Assistance, Inc." })
	end
	ib_action :showAbout # First Cuts > About
	
	def showHelp(sender)
		openPDF("First Cuts for FCP - Help.pdf")
	end
	ib_action :showHelp # Help > First Cuts Help

	def showHelp2(sender)
		openPDF("Finisher for FCP - Help.pdf")
	end
	ib_action :showHelp2 # Help > Finisher Help

	def showHelp3(sender)
		openPDF("exceLogger - Help.pdf")
	end
	ib_action :showHelp3 # Help > exceLogger Help
		
	def showCheatSheet(sender)
		openPDF("Logging \"Cheat Sheet\".pdf")
	end
	ib_action :showCheatSheet # Help > Logging "Cheat Sheet"
	
	def openPDF(filename)
		path_to_file = __FILE__.sub("AppController.rb",filename)
		system('open', path_to_file)
	end

	def visitWebsite(sender)
		system('open', 'http://www.AssistedEditing.com/')
	end
	ib_action :visitWebsite # Help > Visit www.AssistedEditing.com

	def sendFeedback(sender)
		OSX::JRFeedbackController.showFeedback
	end
	ib_action :sendFeedback # First Cuts > Provide Feedback

	def sendSupportEmail(subject)
		system('open', "mailto:techsupport@intelligentassistance.com?subject=First Cuts for FCP #{subject}&body=Attached is the XML file that First Cuts for FCP was unable to process.")
		# &attachment=\"#{attachment}\"
	end

	def chooseXmlFile
		@firstCutsBrowseFileButton.setEnabled(false)
		@finisherBrowseFileButton.setEnabled(false)
		filetypes = ["xml"]
		oPanel = OSX::NSOpenPanel.openPanel
		oPanel.setTitle("Choose the exported Bin(s) of Clips to process (XML file)")
		oPanel.setAllowsMultipleSelection(false)
		oPanel.beginSheetForDirectory_file_types_modalForWindow_modalDelegate_didEndSelector_contextInfo(nil, nil, filetypes, @mainWindow, self, 'didEndOpenSheet:returnCode:contextInfo:', nil)
	end
	ib_action :chooseXmlFile # File > Open, or click Open button

	def chooseSnapshotFile
		@firstCutsBrowseFileButton.setEnabled(false)
		@finisherBrowseFileButton.setEnabled(false)
		filetypes = ["fcf"]
		oPanel = OSX::NSOpenPanel.openPanel
		oPanel.setTitle("Choose the Snapshot to open")
		oPanel.setAllowsMultipleSelection(false)
		oPanel.beginSheetForDirectory_file_types_modalForWindow_modalDelegate_didEndSelector_contextInfo(nil, nil, filetypes, @mainWindow, self, 'didEndOpenSheet:returnCode:contextInfo:', nil)
	end
	ib_action :chooseSnapshotFile # File > Restore Snapshot

	def didEndOpenSheet_returnCode_contextInfo(oPanel, buttonClicked, info)
		if buttonClicked == OSX::NSOKButton
			oPanel.orderOut(self)
			file = oPanel.filenames.objectAtIndex(0)
			application_openFile(self, file)
		else
			@firstCutsBrowseFileButton.setEnabled(true)
			@finisherBrowseFileButton.setEnabled(true)
		end
	end

	def application_openFile(sender, path)
		OSX::NSDocumentController.sharedDocumentController().noteNewRecentDocumentURL( OSX::NSURL.fileURLWithPath(path) ) # add this file to the Recent menu
		clearParamChoices()
		begin
			fileManager = OSX::NSFileManager.defaultManager
			# file_contents_string = OSX::NSString.alloc.initWithData_encoding(fileManager.contentsAtPath(path), OSX::NSUTF8StringEncoding)
			@getInfoMenuItem.setEnabled(false)
			@saveSnapshotMenuItem.setEnabled(false)
			@input_bin_file = path.to_s
			@extension = File.extname(@input_bin_file)
			if @extension==".xml"
				displayProgressSheet()
				@input_bin_file_name = File.basename(@input_bin_file, ".xml")
				@progressMessage.setStringValue("Opening \"" + @input_bin_file_name + "\"...")
				@progressIndicator.setIndeterminate(true)
				@progressIndicator.startAnimation(self)
				# parse_xml_bins(file_contents_string)
				parse_xml_bins(@input_bin_file)
				return true
			elsif @extension==".fcf"
				puts "unpacking a .fcf file..."
				@input_bin_file_name = File.basename(@input_bin_file, ".fcf")
				@getInfoMenuItem.setEnabled(true)
				@saveSnapshotMenuItem.setEnabled(true)
				unpack_snapshot(@input_bin_file)
				return true
			else
				return false
			end
		rescue
			didEndProgressSheet_returnCode_contextInfo(@progressWindow, -1, nil)
			alert = OSX::NSAlert.alloc.init
			alert.setAlertStyle(OSX::NSInformationalAlertStyle) # or NSCriticalAlertStyle
			alert.addButtonWithTitle("Email Tech Support")
			alert.addButtonWithTitle("Ignore")
			if @extension==".xml"
				alert.setMessageText("There was a problem with this XML file")
				puts "RESCUE!!! Problem with parse_xml_bins()"
				alert.setInformativeText("Please choose another XML file exported from Final Cut Pro.\n\nWould you like to email this XML file to Tech Support?")
				buttonClicked = alert.runModal
				sendSupportEmail("problem reading XML file") if buttonClicked==OSX::NSAlertFirstButtonReturn
				chooseXmlFile()
			else
				alert.setMessageText("There was a problem with this Snapshot file")
				puts "RESCUE!!! Problem with unpack_snapshot(shapshot_file)"
				alert.setInformativeText("There was a problem with this file.\n\nPlease choose another Snapshot file.")
				buttonClicked = alert.runModal
				chooseSnapshotFile()
			end
			return false
		end
	end

	def displayProgressSheet
		OSX::NSApp.beginSheet_modalForWindow_modalDelegate_didEndSelector_contextInfo( @progressWindow, @mainWindow, self, 'didEndProgressSheet:returnCode:contextInfo:', nil )
	end
	
	def didEndProgressSheet_returnCode_contextInfo(sheet, code, info)
		sheet.orderOut(nil) # or self?
		@progressMessage.setStringValue("")
		@progressIndicator.stopAnimation(self)
		@progressIndicator.setIndeterminate(true)
		OSX::NSApp.endSheet(sheet)
		if code==0 # we've finished parsing the XML file
			puts "PARSER"
			@saveSnapshotMenuItem.setEnabled(true)
			checkForInconsistencies()
		elsif code==1 # we've finished building the XML sequence to save
			puts "CREATOR"
			saveXmlFile()
			@previewArollButton.setEnabled(true)
			@startTheEditButton.setEnabled(true) if @this_rego[0]["active"]
		elsif code==2 # we've finished adding tracks to the XML sequence to save
			puts "ADDER"
			saveXmlFile()
			@finishTheEditButton.setEnabled(true) if @this_rego[0]["active"]
		end
	end

	def checkForInconsistencies
		inconsistency = OSX::NSAlert.alloc.init
		inconsistency.setAlertStyle(OSX::NSInformationalAlertStyle)
		if @toolbar.selectedItemIdentifier==TOOLBAR_FIRST_IDENTIFIER and @has_sequence==true
			# we're in First Cuts view, but this XML has a sequence - ask if you want to switch to Finisher or ignore the sequence
			inconsistency.setMessageText("The exported XML file has a Sequence")
			inconsistency.setInformativeText("Would you like to switch to Finisher?")
			inconsistency.addButtonWithTitle("OK")
			inconsistency.addButtonWithTitle("Cancel")
			inconsistency.beginSheetModalForWindow_modalDelegate_didEndSelector_contextInfo( @mainWindow, self, "didEndSwitchToFinisherSheet:returnCode:contextInfo:", nil )
		elsif @toolbar.selectedItemIdentifier==TOOLBAR_SECOND_IDENTIFIER and @has_sequence==false
			# we're in Finisher view, but there's no sequence - ask if you want to switch to First Cuts
			inconsistency.setMessageText("The exported XML file has no Sequences")
			inconsistency.setInformativeText("Would you like to switch to First Cuts?")
			inconsistency.addButtonWithTitle("OK")
			inconsistency.addButtonWithTitle("Cancel")
			inconsistency.beginSheetModalForWindow_modalDelegate_didEndSelector_contextInfo( @mainWindow, self, "didEndSwitchToFirstCutsSheet:returnCode:contextInfo:", nil )
		else # no inconsistencies
			build_report_message()
			setupParamChoices()
		end
	end

	def didEndSwitchToFinisherSheet_returnCode_contextInfo(sheet, buttonClicked, info)
		sheet.window.orderOut(nil)
		if buttonClicked == OSX::NSAlertFirstButtonReturn # OK we'll switch to Finisher
		else # stay in First Cuts and ignore the sequence
			@has_sequence = false
		end
		build_report_message()
		setupParamChoices()
	end

	def didEndSwitchToFirstCutsSheet_returnCode_contextInfo(sheet, buttonClicked, info)
		sheet.window.orderOut(nil)
		if buttonClicked == OSX::NSAlertFirstButtonReturn # OK we'll switch to First Cuts
			build_report_message()
			setupParamChoices()
		else
			@finisherBrowseFileButton.setEnabled(true)
		end
	end

	def displayReportSheet
		@getInfoMenuItem.setEnabled(true)
		report = OSX::NSAlert.alloc.init
		if @report_alert==false
			report.setMessageText("\"" + @input_bin_file_name + "\" contains:")
			report.setAlertStyle(OSX::NSInformationalAlertStyle)
		else
			report.setMessageText("There may not be enough information in the log notes to produce a useful edit.\n\n\"" + @input_bin_file_name + "\" contains:")
			report.setAlertStyle(OSX::NSCriticalAlertStyle)
		end
		report.setInformativeText(@report_message)
		report.addButtonWithTitle("OK")
		report.beginSheetModalForWindow_modalDelegate_didEndSelector_contextInfo( @mainWindow, self, "didEndReportSheet:returnCode:contextInfo:", nil )
	end
	ib_action :displayReportSheet # File > Get Info

	def didEndReportSheet_returnCode_contextInfo(sheet, code, info)
		if @has_sequence==false
			showFirstCutsView()
		else
			showFinisherView()
		end
	end

	def clearParamChoices
		@infoTextField.setHidden(true)
		@firstCutsBrowseFileButton.setEnabled(false)
		@finisherBrowseFileButton.setEnabled(false)
		# Opener popup menu
		@openerPopup.removeAllItems
		@openerPopup.setEnabled(false)
		# Lower Thirds popup menu
		@firstCutsLowerThirdPopup.removeAllItems
		@firstCutsLowerThirdPopup.setEnabled(false)
		@finisherLowerThirdPopup.removeAllItems
		@finisherLowerThirdPopup.setEnabled(false)
		# Sequence Settings box
		@frameRatePopup.removeAllItems
		@frameRatePopup.setEnabled(false)
		@frameSizePopup.removeAllItems
		@frameSizePopup.setEnabled(false)
		@aspectRatioPopup.removeAllItems
		@aspectRatioPopup.setEnabled(false)
		@anamorphicCheckbox.setEnabled(false)
		# Story Keywords table
		@story_keywords_available = []
		@story_keywords_chosen = []
		@keywordsAvailableTableView.setEnabled(false)
		@keywordsChosenTableView.setEnabled(false)
		@keywordsAvailableTableView.reloadData
		@keywordsChosenTableView.reloadData
		# Duration text field
		@durationRadioButtons.setEnabled(false)
		@durationTextField.setEnabled(false)
		@durationTextField.setStringValue("")
		@finishTheEditButton.setEnabled(false)
		@previewArollButton.setEnabled(false)
		@startTheEditButton.setEnabled(false)
	end

	def setupParamChoices
		# all interface elements were disabled with clearParamChoices - just enable the ones you need
		@firstCutsBrowseFileButton.setEnabled(true)
		@finisherBrowseFileButton.setEnabled(true)
		puts "@has_sequence ? #{@has_sequence}"
		if @has_sequence==false
			@infoTextField.setStringValue("There are about #{@aroll_shots_total_duration} minutes of A-roll shots available.")
			@infoTextField.setHidden(false)
			# Opener popup menu
			unless @opener_clips==[]
				@openerPopup.removeAllItems
				opener_choices = Array.new
				@opener_clips = @opener_clips.sort_by { |clip| clip[:name] }
				@opener_clips.each do |clip|
					clip_name = clip[:name].split('.')
					opener_choices << clip_name[0]
				end
				opener_choices.each do |title|
					@openerPopup.addItemWithTitle(title)
				end
				@openerPopup.setEnabled(true)
				@openerPopup.selectItemAtIndex(@openerPopup_selected_index)
			end
		else
			@finishTheEditButton.setEnabled(true) if @this_rego[0]["active"]
			@infoTextField.setHidden(true)
			@openerPopup.removeAllItems
			@openerPopup.setEnabled(false)
		end
		# Lower Thirds popup menu
		@firstCutsLowerThirdPopup.removeAllItems
		@finisherLowerThirdPopup.removeAllItems
		lower_third_choices = Array.new
		unless @lower_third_clips==[]
			@lower_third_clips = @lower_third_clips.sort_by { |clip| clip[:name] }
			@lower_third_clips.each do |clip|
				clip_name = clip[:name].split('.')
				lower_third_choices << clip_name[0]
			end
		end
		lower_third_choices << "default"
		lower_third_choices.each do |title|
			@firstCutsLowerThirdPopup.addItemWithTitle(title)
			@finisherLowerThirdPopup.addItemWithTitle(title)
		end
		if @anamorphic.length > 1 # there's more than one anamorphic setting to choose from
			@anamorphicCheckbox.setEnabled(true)
		end
		if @has_sequence==true
			@finisherLowerThirdPopup.setEnabled(true)
			@finisherLowerThirdPopup.selectItemAtIndex(@lowerThirdPopup_selected_index)
			@finishTheEditButton.setEnabled(true) if @this_rego[0]["active"]
		else # don't need sequence settings, keywords or duration for edit valet
			@firstCutsLowerThirdPopup.setEnabled(true)
			@firstCutsLowerThirdPopup.selectItemAtIndex(@lowerThirdPopup_selected_index)
			# Sequence Settings box
			@frameRatePopup.removeAllItems
			@clip_frame_rates.each do |item|
				@frameRatePopup.addItemWithTitle(item[:type])
			end
			@frameRatePopup.setEnabled(true)
			@frameSizePopup.removeAllItems
			@width_height.each do |item|
				@frameSizePopup.addItemWithTitle(item)
			end
			@frameSizePopup.setEnabled(true)
			@aspectRatioPopup.removeAllItems
			@pixelaspectratio.each do |item|
				@aspectRatioPopup.addItemWithTitle(item)
			end
			@aspectRatioPopup.setEnabled(true)
			if @anamorphic[0]=="TRUE"
				@anamorphicCheckbox.setState(1)
			else
				@anamorphicCheckbox.setState(0)
			end
			# Story Keywords table
			@story_keywords_available = @story_keywords - @story_keywords_chosen
			@story_keywords_available = @story_keywords_available.sort_by { |item| item["keyword"] }
			# puts "story keywords #{@story_keywords.inspect}"
			@keywordsAvailableTableView.setEnabled(true)
			@keywordsChosenTableView.setEnabled(true)
			@keywordsAvailableTableView.reloadData
			@keywordsChosenTableView.reloadData
			sortDescriptor = Array.new
			sortDescriptor << OSX::NSSortDescriptor.alloc.initWithKey_ascending_selector( "keyword", true, "localizedCaseInsensitiveCompare:" )
			@keywordsAvailableTableView.setSortDescriptors(sortDescriptor)
			# Duration text field
			@durationRadioButtons.setEnabled(true)
			@durationRadioButtons.selectCellWithTag(@durationRadioButtons_state)
			@durationTextField.setEnabled(true)
			@durationTextField.setStringValue(@durationTextField_value)
			puts "finished length = #{@durationTextField_value} mins"
			# @infoTextField.setStringValue("The A-roll sequence has #{@aroll.length} Clips.")
			@infoTextField.setHidden(false)
		end
	end

	def saveSnapshot
		sPanel = OSX::NSSavePanel.savePanel
		sPanel.setTitle("Save the Snapshot file")
		sPanel.setRequiredFileType("fcf")
		sPanel.setExtensionHidden(false)
		sPanel.setCanSelectHiddenExtension(true)
		sPanel.beginSheetForDirectory_file_modalForWindow_modalDelegate_didEndSelector_contextInfo(nil, nil, @mainWindow, self, 'didEndSaveSnapshot:returnCode:contextInfo:', nil)
	end
	ib_action :saveSnapshot # File > Save Snapshot

	def didEndSaveSnapshot_returnCode_contextInfo(sPanel, buttonClicked, info)
		########## SAVE SNAPSHOT ##########
		if buttonClicked == OSX::NSOKButton
			sPanel.orderOut(self)
			if sPanel.isExtensionHidden?
				extension_hidden = 1
			else
				extension_hidden = 0
			end
			puts "is extension hidden?", sPanel.isExtensionHidden
			snapshot_file = sPanel.filenames.objectAtIndex(0).to_s
			pack_snapshot(snapshot_file)
			fileManager = OSX::NSFileManager.defaultManager
			fileAttributes = { :NSFileExtensionHidden => extension_hidden }
			fileManager.setAttributes_ofItemAtPath_error(fileAttributes, snapshot_file, nil)
		end
	end

	def addKeyword(sender)
		@keywordsAvailableTableView.selectedRowIndexes.to_a.each do |row|
			@story_keywords_chosen << @story_keywords_available[row]
		end
		@story_keywords_available = @story_keywords - @story_keywords_chosen
		# @story_keywords_available = @story_keywords_available.sort_by { |item| item["keyword"] }
		tableView_sortDescriptorsDidChange(@keywordsAvailableTableView, nil)
		@keywordsAvailableTableView.deselectAll(nil)
		@keywordsChosenTableView.deselectAll(nil)
		@keywordsAvailableTableView.reloadData
		@keywordsChosenTableView.reloadData
	end
	ib_action :addKeyword # on click of Add button

	def randomizeKeywords(sender)
		@story_keywords_chosen = @story_keywords_chosen.sort_by {rand}
		@keywordsChosenTableView.deselectAll(nil)
		@keywordsChosenTableView.reloadData
	end
	ib_action :randomizeKeywords # on click of Randomize button

	def removeKeyword(sender)
		deleted_keywords = Array.new
		@keywordsChosenTableView.selectedRowIndexes.to_a.each do |row|
			@story_keywords_available << @story_keywords_chosen[row]
			deleted_keywords << @story_keywords_chosen[row]
		end
		# @story_keywords_available = @story_keywords_available.sort_by { |item| item["keyword"] }
		tableView_sortDescriptorsDidChange(@keywordsAvailableTableView, nil)
		@story_keywords_chosen -= deleted_keywords
		@removeButton.setEnabled(false) # bugfix?
		@keywordsAvailableTableView.deselectAll(nil)
		@keywordsChosenTableView.deselectAll(nil)
		@keywordsAvailableTableView.reloadData
		@keywordsChosenTableView.reloadData
	end
	ib_action :removeKeyword # on click of Remove button
	
	def removeAllKeywords(sender)
		@story_keywords_available = @story_keywords
		# @story_keywords_available = @story_keywords_available.sort_by { |item| item["keyword"] }
		tableView_sortDescriptorsDidChange(@keywordsAvailableTableView, nil)
		@story_keywords_chosen = []
		@removeButton.setEnabled(false)
		@keywordsAvailableTableView.deselectAll(nil)
		@keywordsChosenTableView.deselectAll(nil)
		@keywordsAvailableTableView.reloadData
		@keywordsChosenTableView.reloadData
	end
	ib_action :removeAllKeywords # on click of Remove All button

	def previewStory(sender)
		@firstCutsSpinner.startAnimation(self)
		if @durationRadioButtons.selectedCell.tag==1 # get duration radio button
			@finished_length = 0 # no limit on duration
		elsif @durationTextField.stringValue.to_s==""
			@finished_length = 0
		else
			@finished_length = (@durationTextField.stringValue.to_s).to_f # get duration
		end
		@chosen_lower_third = @lower_third_clips[@firstCutsLowerThirdPopup.indexOfSelectedItem] # get lower third popup
		chosen_frame_rate = FRAME_RATES.find { |rate| rate[:type] == @frameRatePopup.titleOfSelectedItem }
		@fps = chosen_frame_rate[:timebase]
		# get themes into @keyword array
		@keyword = Array.new
		@story_keywords_chosen.each do |keyword|
			@keyword << keyword["keyword"]
		end
		puts @keyword.inspect
		begin
			edit_process("video")
			play_story()
			@firstCutsSpinner.stopAnimation(self)
			alert = OSX::NSAlert.alloc.init
			alert.setAlertStyle(OSX::NSInformationalAlertStyle)
			alert.addButtonWithTitle("OK")
			if @result=="no fcp_rate"
				alert.setMessageText("Preview is not available for older snapshots")
				alert.setInformativeText("Please import the Final Cut Pro XML again.")
				buttonClicked = alert.runModal
			elsif @result=="no movie"
				alert.setMessageText("Preview is not available for offline files")
				alert.setInformativeText("Please reconnect media in Final Cut Pro and export XML again.")
				buttonClicked = alert.runModal
			end
		rescue
			puts "RESCUE!!! Problem with edit_process()"
			@firstCutsSpinner.stopAnimation(self)
			alert = OSX::NSAlert.alloc.init
			alert.setAlertStyle(OSX::NSInformationalAlertStyle) # or NSCriticalAlertStyle
			alert.setMessageText("There was a problem previewing this edit")
			alert.setInformativeText("Please try again using different keywords or a different exported file.\n")
			alert.addButtonWithTitle("OK")
			buttonClicked = alert.runModal
			@previewArollButton.setEnabled(true)
			@startTheEditButton.setEnabled(true) if @this_rego[0]["active"]
		end
	end
	ib_action :previewStory

	def startTheEdit(sender)
		@firstCutsSpinner.startAnimation(self)
		@startTheEditButton.setEnabled(false)
		@startTheEditButton.displayIfNeeded
		if @durationRadioButtons.selectedCell.tag==1 # get duration radio button
			@finished_length = 0 # no limit on duration
		elsif @durationTextField.stringValue.to_s==""
			@finished_length = 0
		else
			@finished_length = (@durationTextField.stringValue.to_s).to_f # get duration
		end
		@chosen_opener = @opener_clips[@openerPopup.indexOfSelectedItem] # get opener popup
		if @chosen_opener.nil?
			@chosen_opener = ""
		end
		@chosen_lower_third = @lower_third_clips[@firstCutsLowerThirdPopup.indexOfSelectedItem] # get lower third popup
		puts "******* @chosen_lower_third = #{@chosen_lower_third.inspect}"
		chosen_frame_rate = FRAME_RATES.find { |rate| rate[:type] == @frameRatePopup.titleOfSelectedItem }
		@ntsc = chosen_frame_rate[:ntsc]
		@fps = chosen_frame_rate[:timebase]
		chosen_frame_size = @frameSizePopup.titleOfSelectedItem.split(" × ")
		@width = chosen_frame_size[0].to_s # fussy REXML needs these to be strings
		@height = chosen_frame_size[1].to_s
		@pixel_aspect_ratio = @aspectRatioPopup.titleOfSelectedItem.to_s
		if @anamorphicCheckbox.state==1 # get anamorphic checkbox
			@anamorphic_state = "TRUE"
		else
			@anamorphic_state = "FALSE"
		end
		puts "#{@ntsc} #{@fps}fps, #{@width}x#{@height}, #{@pixel_aspect_ratio}, #{@anamorphic_state}"
		# get themes into @keyword array
		@keyword = Array.new
		@story_keywords_chosen.each do |keyword|
			@keyword << keyword["keyword"]
		end
		puts @keyword.inspect
		begin
			edit_process("xml")
			########## BUILD RESULTS ##########
			@builder = false
			if @keyword.length==1
				suggested_filename = @keyword[0]
			else
				suggested_filename = @keyword[0] + ", etc"
			end
			@firstCutsSpinner.stopAnimation(self)
			sPanel = OSX::NSSavePanel.savePanel
			sPanel.setTitle("Save the edited sequence XML file")
			sPanel.setRequiredFileType("xml")
			sPanel.setExtensionHidden(true)
			sPanel.setCanSelectHiddenExtension(false)
			sPanel.beginSheetForDirectory_file_modalForWindow_modalDelegate_didEndSelector_contextInfo(nil, suggested_filename, @mainWindow, self, 'didEndSaveSheet:returnCode:contextInfo:', nil)
		rescue
			puts "RESCUE!!! Problem with edit_process()"
			@firstCutsSpinner.stopAnimation(self)
			alert = OSX::NSAlert.alloc.init
			alert.setAlertStyle(OSX::NSInformationalAlertStyle) # or NSCriticalAlertStyle
			alert.setMessageText("There was a problem creating this edit")
			alert.setInformativeText("Please try again using different keywords or a different exported file.\n\nWould you like to email this XML file to Tech Support?")
			alert.addButtonWithTitle("Email Tech Support")
			alert.addButtonWithTitle("Ignore")
			buttonClicked = alert.runModal
			sendSupportEmail("problem creating a First Cuts edit") if buttonClicked==OSX::NSAlertFirstButtonReturn
			@previewArollButton.setEnabled(true)
			@startTheEditButton.setEnabled(true) if @this_rego[0]["active"]
		end
	end
	ib_action :startTheEdit # on click of Make the Edit button

	def finishTheEdit(sender)
		@finishTheEditButton.setEnabled(false)
		@chosen_opener = "" #@opener_clips[@openerPopup.indexOfSelectedItem] # get opener popup
		@chosen_lower_third = @lower_third_clips[@finisherLowerThirdPopup.indexOfSelectedItem] # get lower third popup
		puts "#{@ntsc} #{@fps}fps, #{@width}x#{@height}, #{@pixel_aspect_ratio}, #{@anamorphic_state}"
		begin
			edit_valet_process()
			puts "@suggested_filename #{@suggested_filename}"
			########## BUILD RESULTS ##########
			@builder = false
			sPanel = OSX::NSSavePanel.savePanel
			sPanel.setTitle("Save the finished sequence XML file")
			sPanel.setRequiredFileType("xml")
			sPanel.setExtensionHidden(true)
			sPanel.setCanSelectHiddenExtension(false)
			sPanel.beginSheetForDirectory_file_modalForWindow_modalDelegate_didEndSelector_contextInfo(nil, @suggested_file_name, @mainWindow, self, 'didEndSaveSheet:returnCode:contextInfo:', nil)
		rescue
			puts "RESCUE!!! Problem with edit_valet_process()"
			alert = OSX::NSAlert.alloc.init
			alert.setAlertStyle(OSX::NSInformationalAlertStyle) # or NSCriticalAlertStyle
			alert.setMessageText("There was a problem finishing this edit")
			alert.setInformativeText("Please try again using a different exported file.\n\nWould you like to email this XML file to Tech Support?")
			alert.addButtonWithTitle("Email Tech Support")
			alert.addButtonWithTitle("Ignore")
			buttonClicked = alert.runModal
			sendSupportEmail("problem finishing a sequence") if buttonClicked==OSX::NSAlertFirstButtonReturn
		end
	end
	ib_action :finishTheEdit # on click of Finish the Edit button

	def didEndSaveSheet_returnCode_contextInfo(sPanel, buttonClicked, info)
		########## SAVE RESULTS ##########
		if buttonClicked == OSX::NSOKButton
			sPanel.orderOut(self)
			displayProgressSheet()
			@progressMessage.setStringValue("Creating the edited sequence XML file...")
			@progressIndicator.setIndeterminate(true)
			@progressIndicator.startAnimation(self)
			@output_sequence_file = sPanel.filenames.objectAtIndex(0).to_s
			@title = File.basename(@output_sequence_file, ".xml") # make the sequence title the user entered file name
			puts "@has_sequence #{@has_sequence}"
			begin
				if @has_sequence==false
					puts "Save as (using REXML)... #{@output_sequence_file}"
					create_xml()
				elsif @has_sequence==true
					puts "Add tracks (using REXML)..."
					add_tracks_to_xml()
				end
			rescue
				puts "RESCUE!!! Problem creating the file"
				alert = OSX::NSAlert.alloc.init
				alert.setAlertStyle(OSX::NSInformationalAlertStyle) # or NSCriticalAlertStyle
				alert.setMessageText("There was a problem saving this edit")
				alert.setInformativeText("Please try again using different keywords or a different exported file.\n\nWould you like to email this XML file to Tech Support?")
				alert.addButtonWithTitle("Email Tech Support")
				alert.addButtonWithTitle("Ignore")
				buttonClicked = alert.runModal
				sendSupportEmail("problem saving XML file") if buttonClicked==OSX::NSAlertFirstButtonReturn
			end
		else
			if @has_sequence==false
				@previewArollButton.setEnabled(true)
				@startTheEditButton.setEnabled(true) if @this_rego[0]["active"]
			elsif @has_sequence==true
				@finishTheEditButton.setEnabled(true) if @this_rego[0]["active"]
			end
		end
	end

	def saveXmlFile # format and save @sequence_xml to the file @output_sequence_file as a Final Cut Pro XML
		fileManager = OSX::NSFileManager.defaultManager
		fileAttributes = { :NSFileExtensionHidden => 1, :NSFileHFSCreatorCode => 1264941383 } # extension is hidden, Final Cut Pro is creator code
		fileManager.setAttributes_ofItemAtPath_error(fileAttributes, @output_sequence_file, nil)
	end

	### exceLogger methods ###
	
	def chooseExcelFile
		@saveXMLButton.setEnabled(false)
		@excel_file = nil
		@excelIcon.setHidden(true)
		@excelIcon.setToolTip("")
		@excelFileName.setStringValue("")
		filetypes = ["xls","xlsx"]
		oPanel = OSX::NSOpenPanel.openPanel
		oPanel.setTitle("Choose the Excel document to open")
		oPanel.setAllowsMultipleSelection(false)
		oPanel.beginSheetForDirectory_file_types_modalForWindow_modalDelegate_didEndSelector_contextInfo(nil, nil, filetypes, @mainWindow, self, 'didEndLoggerOpenSheet:returnCode:contextInfo:', nil)
	end
	ib_action :chooseExcelFile
	
	def chooseBatchListXmlFile
		@saveXMLButton.setEnabled(false)
		@xml_file = nil
		@xmlIcon.setHidden(true)
		@xmlIcon.setToolTip("")
		@xmlFileName.setStringValue("")
		filetypes = ["xml"]
		oPanel = OSX::NSOpenPanel.openPanel
		oPanel.setTitle("Choose the exported Bin(s) of Clips to process (XML file)")
		oPanel.setAllowsMultipleSelection(false)
		oPanel.beginSheetForDirectory_file_types_modalForWindow_modalDelegate_didEndSelector_contextInfo(nil, nil, filetypes, @mainWindow, self, 'didEndLoggerOpenSheet:returnCode:contextInfo:', nil)
	end
	ib_action :chooseBatchListXmlFile
	
	def didEndLoggerOpenSheet_returnCode_contextInfo(oPanel, buttonClicked, info)
		if buttonClicked == OSX::NSOKButton
			oPanel.orderOut(self)
			file = oPanel.filenames.objectAtIndex(0)
			application_setLoggerFile(self, file)
		end
	end
	
	def application_setLoggerFile(sender, path)
		OSX::NSDocumentController.sharedDocumentController().noteNewRecentDocumentURL( OSX::NSURL.fileURLWithPath(path) ) # add this file to the Recent menu
		fileManager = OSX::NSFileManager.defaultManager
		chosen_file = path.to_s
		chosen_file_extension = File.extname(chosen_file)
		if chosen_file_extension==".xls" || chosen_file_extension==".xlsx"
			@excel_file = chosen_file
			@excelIcon.setHidden(false)
			@excelIcon.setToolTip(path.to_s)
			@excelFileName.setStringValue(File.basename(chosen_file))
		elsif chosen_file_extension==".xml"
			@xml_file = chosen_file
			@xmlIcon.setHidden(false)
			@xmlIcon.setToolTip(path.to_s)
			@xmlFileName.setStringValue(File.basename(chosen_file))
		else
			return false
		end
		if @excel_file!=nil && @xml_file!=nil
			@saveXMLButton.setEnabled(true)
		end
	end
	
	def saveLoggedXml
		@progressIndicator.setIndeterminate(true)
		@progressIndicator.setUsesThreadedAnimation(true)
		@progressIndicator.startAnimation(self)
		@progressMessage.setStringValue("Adding log notes...")
		puts "on save radio buttons:", @onSaveRadioButtons.selectedCell.tag.inspect
		if @onSaveRadioButtons.selectedCell.tag==0 # save XML file
			sPanel = OSX::NSSavePanel.savePanel
			sPanel.setTitle("Save the new XML file")
			sPanel.setRequiredFileType("xml")
			sPanel.setExtensionHidden(false)
			sPanel.setCanSelectHiddenExtension(true)
			sPanel.beginSheetForDirectory_file_modalForWindow_modalDelegate_didEndSelector_contextInfo(nil, nil, @mainWindow, self, 'didEndSaveLoggedXml:returnCode:contextInfo:', nil)
		else # save temp and open into FCP project
			begin
				displayProgressSheet()
				add_log_notes(@excel_file, @xml_file, TEMP_XML_FILE)
				@progressMessage.setStringValue("Switching to Final Cut Pro\n")
				@progressView.displayIfNeeded
				fileManager = OSX::NSFileManager.defaultManager
				fileAttributes = { :NSFileExtensionHidden => 1, :NSFileHFSCreatorCode => 1264941383 } # extension is hidden, Final Cut Pro is creator code
				fileManager.setAttributes_ofItemAtPath_error(fileAttributes, TEMP_XML_FILE, nil)
				sleep 2
				OSX::NSWorkspace.sharedWorkspace.openFile_withApplication(TEMP_XML_FILE, "Final Cut Pro")
				didEndProgressSheet_returnCode_contextInfo(@progressWindow, 2, nil)
			rescue
				didEndProgressSheet_returnCode_contextInfo(@progressWindow, -1, nil)
				alert = OSX::NSAlert.alloc.init
				alert.setMessageText("There was a problem creating this XML file.")
				alert.setAlertStyle(OSX::NSInformationalAlertStyle)
				puts "RESCUE!!! Problem with explore_metadata()"
				alert.setInformativeText("Please choose another XML file exported from\nFinal Cut Pro.")
				alert.addButtonWithTitle("OK")
				alert.runModal
			end
		end
	end
	ib_action :saveLoggedXml
	
	def didEndSaveLoggedXml_returnCode_contextInfo(sPanel, buttonClicked, info)
		########## SAVE LOGGED XML ##########
		if buttonClicked == OSX::NSOKButton
			sPanel.orderOut(self)
			if sPanel.isExtensionHidden?
				extension_hidden = 1
			else
				extension_hidden = 0
			end
			puts "is extension hidden?", sPanel.isExtensionHidden
			new_xml_file = sPanel.filenames.objectAtIndex(0).to_s
			displayProgressSheet()
			begin
				add_log_notes(@excel_file, @xml_file, new_xml_file)
				fileManager = OSX::NSFileManager.defaultManager
				fileAttributes = { :NSFileExtensionHidden => 1, :NSFileHFSCreatorCode => 1264941383 } # extension is hidden, Final Cut Pro is creator code
				fileManager.setAttributes_ofItemAtPath_error(fileAttributes, new_xml_file, nil)
				didEndProgressSheet_returnCode_contextInfo(@progressWindow, 1, nil)
			rescue
				didEndProgressSheet_returnCode_contextInfo(@progressWindow, -1, nil)
				alert = OSX::NSAlert.alloc.init
				alert.setMessageText("There was a problem saving this XML file.")
				alert.setAlertStyle(OSX::NSInformationalAlertStyle)
				puts "RESCUE!!! Problem with explore_metadata()"
				alert.setInformativeText("Please choose another XML file exported from\nFinal Cut Pro.")
				alert.addButtonWithTitle("OK")
				alert.runModal
			end
		end
	end

	### end exceLogger methods ###	
	def applicationShouldTerminateAfterLastWindowClosed(sender)
		true
	end

	### Toolbar methods ###
	
	def setupToolbar
		# Create a new toolbar instance, and attach it to our document window.
		@toolbar ||= OSX::NSToolbar.alloc.initWithIdentifier(TOOLBAR_IDENTIFIER)
		# Set up toolbar properties: disallow customization, give a default display mode and don't remember state in user defaults.
		@toolbar.setAllowsUserCustomization(false)
		@toolbar.setAutosavesConfiguration(false)
		@toolbar.setDisplayMode(OSX::NSToolbarDisplayModeIconAndLabel)
		# We are the delegate.
		@toolbar.setDelegate(self)
		# Attach the toolbar to the window.
		@mainWindow.setToolbar(@toolbar)
	end

	def toolbarDefaultItemIdentifiers(toolbar)
		[ TOOLBAR_FIRST_IDENTIFIER, TOOLBAR_SECOND_IDENTIFIER, TOOLBAR_THIRD_IDENTIFIER ]
	end

	def toolbarAllowedItemIdentifiers(toolbar)
		[ TOOLBAR_FIRST_IDENTIFIER, TOOLBAR_SECOND_IDENTIFIER, TOOLBAR_THIRD_IDENTIFIER ]
	end
	
	def toolbarSelectableItemIdentifiers(toolbar)
		[ TOOLBAR_FIRST_IDENTIFIER, TOOLBAR_SECOND_IDENTIFIER, TOOLBAR_THIRD_IDENTIFIER ]
	end
	
	# don't need validateToolbarItem if all items will be shown at all times
	
	def toolbar_itemForItemIdentifier_willBeInsertedIntoToolbar(toolbar, identifier, flag)
		label, tooltip, image, action = 
		case identifier.to_s
		when TOOLBAR_FIRST_IDENTIFIER
			[ "First Cuts", "First Cuts", "FirstCutsToolbarImage", 'showFirstCuts:' ]
		when TOOLBAR_SECOND_IDENTIFIER
			[ "Finisher", "Finisher", "FinisherToolbarImage", 'showFinisher:' ]
		when TOOLBAR_THIRD_IDENTIFIER
			[ "exceLogger", "exceLogger", "exceLoggerToolbarImage", 'showExceLogger:']
		else
			return nil
		end
		toolbarItem = OSX::NSToolbarItem.alloc.initWithItemIdentifier(identifier)
		toolbarItem.setLabel(label)
		toolbarItem.setPaletteLabel(label)
		toolbarItem.setToolTip(tooltip)
		toolbarItem.setImage(OSX::NSImage.imageNamed(image))
		toolbarItem.setTarget(self)
		toolbarItem.setAction(action)
		return toolbarItem
	end
	
	### TableView methods ###

	def numberOfRowsInTableView(tableView)
		if tableView.tag==1
			return @story_keywords_available.length
		elsif tableView.tag==2
			if @story_keywords_chosen.length > 2
				@randomizeButton.setEnabled(true)
				@removeAllButton.setEnabled(true)
				@previewArollButton.setEnabled(true)
				@startTheEditButton.setEnabled(true) if @this_rego[0]["active"]
			elsif @story_keywords_chosen.length > 0
				@removeAllButton.setEnabled(true)
				@previewArollButton.setEnabled(true)
				@startTheEditButton.setEnabled(true) if @this_rego[0]["active"]
			else
				@randomizeButton.setEnabled(false) unless @randomizeButton.nil?
				@removeAllButton.setEnabled(false) unless @removeAllButton.nil?
				@previewArollButton.setEnabled(false) unless @previewArollButton.nil?
				@startTheEditButton.setEnabled(false) unless @startTheEditButton.nil?
			end
			return @story_keywords_chosen.length
		end
	end

	def tableView_objectValueForTableColumn_row(tableView, aTableColumn, rowIndex)
		if tableView.tag==1
			@addButton.setEnabled(false)
			if tableView.numberOfSelectedRows > 0
				@addButton.setEnabled(true)
			end
			# return @story_keywords_available[rowIndex]
			@columns.each do |column|
			    if aTableColumn.headerCell.stringValue == column
						object = @story_keywords_available[rowIndex]
						return object[column]
				end
			end
		elsif tableView.tag==2
			@removeButton.setEnabled(false)
			if tableView.numberOfSelectedRows > 0
				@removeButton.setEnabled(true)
			end
			# return @story_keywords_chosen[rowIndex]
			@columns.each do |column|
			    if aTableColumn.headerCell.stringValue == column
					object = @story_keywords_chosen[rowIndex]
					return object[column]
				end
			end
		end
	end

	def tableView_sortDescriptorsDidChange(tableView, oldDescriptors)
		key = tableView.sortDescriptors[0].key.to_s # the first descriptor is always the one used to sort
		ascending = tableView.sortDescriptors[0].ascending
		puts "#{key.inspect} #{ascending.inspect}"
		puts tableView.sortDescriptors.inspect
		if ascending==true
			@story_keywords_available = @story_keywords_available.sort_by { |item| item[key] }
		else
			@story_keywords_available = (@story_keywords_available.sort_by { |item| item[key] }).reverse
		end
		@keywordsAvailableTableView.deselectAll(nil)
		tableView.reloadData
	end

	def tableView_writeRowsWithIndexes_toPasteboard(tableView, rowIndexes, pasteboard)
		if tableView.tag==1
			# puts "Available table: #{rowIndexes.inspect} #{pasteboard.inspect}"
			data = OSX::NSKeyedArchiver.archivedDataWithRootObject rowIndexes
			pasteboard.declareTypes_owner [KeywordsAvailablePasteboard], self
			pasteboard.setData_forType data, KeywordsAvailablePasteboard
			return true
		elsif tableView.tag==2
			# puts "tableView_writeRowsWithIndexes_toPasteboard: #{tableView.tag}, #{rowIndexes.inspect}, #{pasteboard.inspect}"
			data = OSX::NSKeyedArchiver.archivedDataWithRootObject rowIndexes
			pasteboard.declareTypes_owner [KeywordsChosenPasteboard], self
			pasteboard.setData_forType data, KeywordsChosenPasteboard
			return true
		end
	end

	def tableView_validateDrop_proposedRow_proposedDropOperation(tableView, draggingInfo, row, operation)
		if operation==OSX::NSTableViewDropAbove
			return OSX::NSDragOperationEvery
		else
			return OSX::NSDragOperationNone
		end
	end

	def tableView_acceptDrop_row_dropOperation(tableView, draggingInfo, row, operation)
		puts "tableView_acceptDrop_row_dropOperation #{tableView.tag.inspect}, #{draggingInfo.inspect}, #{row.inspect}, #{operation.inspect}"
		pasteboard = draggingInfo.draggingPasteboard
		if tableView.tag==2 and pasteboard.types.containsObject? KeywordsChosenPasteboard # dragging within @keywordsChosenTableView
			data = OSX::NSKeyedUnarchiver.unarchiveObjectWithData( pasteboard.dataForType(KeywordsChosenPasteboard) ).to_a
			puts "moving rows #{data.inspect}"
			data.reverse_each do |index|
				move_me = @story_keywords_chosen[index]
				@story_keywords_chosen[index] = nil
				@story_keywords_chosen.insert(row, move_me)
			end
			@story_keywords_chosen.compact!
			@keywordsChosenTableView.deselectAll(nil)
			tableView.reloadData
			return true
		elsif tableView.tag==2 and pasteboard.types.containsObject? KeywordsAvailablePasteboard # dragging from @keywordsAvailableTableView to @keywordsChosenTableView
			data = OSX::NSKeyedUnarchiver.unarchiveObjectWithData( pasteboard.dataForType(KeywordsAvailablePasteboard) ).to_a
			puts "adding rows #{data.inspect}"
			data.reverse_each do |index|
				move_me = @story_keywords_available[index]
				@story_keywords_available[index] = nil
				@story_keywords_chosen.insert(row, move_me)
			end
			@story_keywords_available.compact!
			@keywordsAvailableTableView.deselectAll(nil)
			@keywordsChosenTableView.deselectAll(nil)
			@keywordsAvailableTableView.reloadData
			@keywordsChosenTableView.reloadData
			return true
		elsif tableView.tag==1 # dragging from @keywordsChosenTableView to @keywordsAvailableTableView
			data = OSX::NSKeyedUnarchiver.unarchiveObjectWithData( pasteboard.dataForType(KeywordsChosenPasteboard) ).to_a
			puts "moving rows #{data.inspect}"
			data.reverse_each do |index|
				move_me = @story_keywords_chosen[index]
				@story_keywords_chosen[index] = nil
				@story_keywords_available.insert(row, move_me)
			end
			# @story_keywords_available = @story_keywords_available.sort_by { |item| item["keyword"] }
			tableView_sortDescriptorsDidChange(@keywordsAvailableTableView, nil)
			@story_keywords_chosen.compact!
			@keywordsAvailableTableView.deselectAll(nil)
			@keywordsChosenTableView.deselectAll(nil)
			@keywordsAvailableTableView.reloadData
			@keywordsChosenTableView.reloadData
			return true
		else
			return false
		end
	end

end
